@@ -1,10 +1,10 @@
-use crate::analyzer::analyze;
 use crate::analyzer::ProgramData;
 use crate::analyzer::UsageAnalyzer;
+use crate::marks::Marks;
 use crate::option::CompressOptions;
 use crate::util::contains_leaping_yield;
+use crate::util::MoudleItemExt;
 use fxhash::FxHashMap;
-use fxhash::FxHashSet;
 use retain_mut::RetainMut;
 use std::fmt::Write;
 use std::mem::take;
@@ -13,7 +13,9 @@ use swc_atoms::JsWord;
 use swc_common::comments::Comments;
 use swc_common::iter::IdentifyLast;
 use swc_common::pass::Repeated;
+use swc_common::sync::Lrc;
 use swc_common::Mark;
+use swc_common::SourceMap;
 use swc_common::Spanned;
 use swc_common::SyntaxContext;
 use swc_common::DUMMY_SP;
@@ -24,6 +26,8 @@ use swc_ecma_utils::undefined;
 use swc_ecma_utils::ExprExt;
 use swc_ecma_utils::ExprFactory;
 use swc_ecma_utils::Id;
+use swc_ecma_utils::IsEmpty;
+use swc_ecma_utils::ModuleItemLike;
 use swc_ecma_utils::StmtLike;
 use swc_ecma_utils::Type;
 use swc_ecma_utils::Value;
@@ -41,12 +45,14 @@ mod computed_props;
 mod conditionals;
 mod dead_code;
 mod evaluate;
+mod fns;
 mod hoist_props;
 mod if_return;
 mod iife;
 mod inline;
 mod join_vars;
 mod loops;
+mod misc;
 mod numbers;
 mod ops;
 mod properties;
@@ -61,8 +67,11 @@ const DISABLE_BUGGY_PASSES: bool = true;
 
 /// This pass is simillar to `node.optimize` of terser.
 pub(super) fn optimizer<'a>(
-    options: CompressOptions,
+    cm: Lrc<SourceMap>,
+    marks: Marks,
+    options: &'a CompressOptions,
     comments: Option<&'a dyn Comments>,
+    data: &'a ProgramData,
 ) -> impl 'a + VisitMut + Repeated {
     assert!(
         options.top_retain.iter().all(|s| s.trim() != ""),
@@ -72,6 +81,8 @@ pub(super) fn optimizer<'a>(
     let done = Mark::fresh(Mark::root());
     let done_ctxt = SyntaxContext::empty().apply_mark(done);
     Optimizer {
+        cm,
+        marks,
         comments,
         changed: false,
         options,
@@ -83,11 +94,10 @@ pub(super) fn optimizer<'a>(
         simple_props: Default::default(),
         _simple_array_values: Default::default(),
         typeofs: Default::default(),
-        data: Default::default(),
+        data: Some(data),
         ctx: Default::default(),
         done,
         done_ctxt,
-        vars_accessible_without_side_effect: Default::default(),
         label: Default::default(),
     }
 }
@@ -100,6 +110,16 @@ struct Ctx {
     /// `true` if the [VarDecl] has const annotation.
     has_const_ann: bool,
 
+    in_bool_ctx: bool,
+
+    in_asm: bool,
+
+    is_callee: bool,
+    in_call_arg: bool,
+
+    /// If this is `true`, the first usage will be inlined as an assignment.
+    inline_as_assignment: bool,
+
     var_kind: Option<VarDeclKind>,
 
     /// `true` if we should not inline values.
@@ -130,6 +150,8 @@ struct Ctx {
     /// `true` while handling inner statements of a labelled statement.
     stmt_lablled: bool,
 
+    dont_use_negated_iife: bool,
+
     /// `true` while handling top-level export decls.
     is_exported: bool,
 
@@ -172,10 +194,14 @@ impl Ctx {
 }
 
 struct Optimizer<'a> {
+    cm: Lrc<SourceMap>,
+
+    marks: Marks,
+
     comments: Option<&'a dyn Comments>,
 
     changed: bool,
-    options: CompressOptions,
+    options: &'a CompressOptions,
 
     /// Statements prepended to the current statement.
     prepend_stmts: Vec<Stmt>,
@@ -196,14 +222,12 @@ struct Optimizer<'a> {
     ///
     /// This is calculated multiple time, but only once per one
     /// `visit_mut_module`.
-    data: Option<ProgramData>,
+    data: Option<&'a ProgramData>,
     ctx: Ctx,
     /// In future: This will be used to `mark` node as done.
     done: Mark,
     done_ctxt: SyntaxContext,
 
-    vars_accessible_without_side_effect: FxHashSet<Id>,
-
     /// Closest label.
     label: Option<Id>,
 }
@@ -221,15 +245,22 @@ impl Repeated for Optimizer<'_> {
 impl Optimizer<'_> {
     fn handle_stmt_likes<T>(&mut self, stmts: &mut Vec<T>)
     where
-        T: StmtLike + VisitMutWith<Self>,
-        Vec<T>: VisitMutWith<Self> + VisitWith<UsageAnalyzer>,
+        T: StmtLike + ModuleItemLike + MoudleItemExt + VisitMutWith<Self>,
+        Vec<T>: VisitMutWith<Self>
+            + VisitWith<UsageAnalyzer>
+            + VisitWith<self::collapse_vars::VarWithOutInitCounter>
+            + VisitMutWith<self::collapse_vars::VarPrepender>
+            + VisitMutWith<self::collapse_vars::VarMover>,
     {
         match self.data {
             Some(..) => {}
             None => {
-                self.data = Some(analyze(stmts));
+                unreachable!()
             }
         }
+        let mut use_asm = false;
+        let prepend_stmts = self.prepend_stmts.take();
+        let append_stmts = self.append_stmts.take();
 
         {
             let mut child_ctx = Ctx { ..self.ctx };
@@ -245,6 +276,12 @@ impl Optimizer<'_> {
                             if v.value == *"use strict" && !v.has_escape {
                                 child_ctx.in_strict = true;
                             }
+
+                            if v.value == *"use asm" && !v.has_escape {
+                                child_ctx.in_asm = true;
+                                self.ctx.in_asm = true;
+                                use_asm = true;
+                            }
                         }
                         _ => {}
                     },
@@ -271,17 +308,33 @@ impl Optimizer<'_> {
             *stmts = new;
         }
 
+        self.ctx.in_asm |= use_asm;
+
+        self.drop_useless_blocks(stmts);
+
+        self.reorder_stmts(stmts);
+
         self.merge_simillar_ifs(stmts);
         self.join_vars(stmts);
 
         self.make_sequences(stmts);
 
-        self.collapse_consequtive_vars(stmts);
+        self.collapse_vars_without_init(stmts);
+
+        self.drop_else_token(stmts);
+
+        self.break_assignments_in_seqs(stmts);
+
+        stmts.extend(self.append_stmts.drain(..).map(T::from_stmt));
 
         stmts.retain(|stmt| match stmt.as_stmt() {
             Some(Stmt::Empty(..)) => false,
             _ => true,
-        })
+        });
+
+        debug_assert_eq!(self.prepend_stmts, vec![]);
+        self.prepend_stmts = prepend_stmts;
+        self.append_stmts = append_stmts;
     }
 
     /// `a = a + 1` => `a += 1`.
@@ -310,6 +363,15 @@ impl Optimizer<'_> {
             _ => return,
         };
 
+        // Don't break code for old browsers.
+        match op {
+            BinaryOp::LogicalOr => return,
+            BinaryOp::LogicalAnd => return,
+            BinaryOp::Exp => return,
+            BinaryOp::NullishCoalescing => return,
+            _ => {}
+        }
+
         let op = match op {
             BinaryOp::In | BinaryOp::InstanceOf => return,
 
@@ -540,7 +602,7 @@ impl Optimizer<'_> {
             }
         }
 
-        log::trace!("Compressing array.join()");
+        log::debug!("Compressing array.join()");
 
         self.changed = true;
         *e = Expr::Lit(Lit::Str(Str {
@@ -580,7 +642,7 @@ impl Optimizer<'_> {
             match lit {
                 Lit::Bool(v) => {
                     self.changed = true;
-                    log::trace!("Compressing boolean literal");
+                    log::debug!("Compressing boolean literal");
                     *e = Expr::Unary(UnaryExpr {
                         span: v.span,
                         op: op!("!"),
@@ -618,15 +680,19 @@ impl Optimizer<'_> {
     /// Returns [None] if expression is side-effect-free.
     /// If an expression has a side effect, only side effects are returned.
     fn ignore_return_value(&mut self, e: &mut Expr) -> Option<Expr> {
+        self.optimize_bang_within_logical_ops(e, true);
+
+        self.compress_cond_to_logical_ignoring_return_value(e);
+
         match e {
             Expr::Ident(..) | Expr::This(_) | Expr::Invalid(_) | Expr::Lit(..) => {
-                log::trace!("ignore_return_value: Dropping unused expr");
+                log::debug!("ignore_return_value: Dropping unused expr");
                 self.changed = true;
                 return None;
             }
             // Function expression cannot have a side effect.
             Expr::Fn(_) => {
-                log::trace!(
+                log::debug!(
                     "ignore_return_value: Dropping unused fn expr as it does not have any side \
                      effect"
                 );
@@ -692,7 +758,12 @@ impl Optimizer<'_> {
                 right,
                 ..
             }) => {
-                let new_r = self.ignore_return_value(right);
+                let ctx = Ctx {
+                    dont_use_negated_iife: self.ctx.dont_use_negated_iife
+                        || self.options.side_effects,
+                    ..self.ctx
+                };
+                let new_r = self.with_ctx(ctx).ignore_return_value(right);
 
                 match new_r {
                     Some(r) => {
@@ -763,15 +834,36 @@ impl Optimizer<'_> {
                 _ => false,
             } && args.is_empty() =>
             {
-                log::trace!("ignore_return_value: Dropping a pure call");
+                log::debug!("ignore_return_value: Dropping a pure call");
                 self.changed = true;
                 return None;
             }
 
+            Expr::Call(CallExpr {
+                callee: ExprOrSuper::Expr(callee),
+                args,
+                ..
+            }) => {
+                if args.is_empty() {
+                    match &mut **callee {
+                        Expr::Fn(f) => {
+                            if f.function.body.is_none()
+                                || f.function.body.as_ref().unwrap().is_empty()
+                            {
+                                return None;
+                            }
+                        }
+                        _ => {}
+                    }
+                }
+
+                return Some(e.take());
+            }
+
             Expr::MetaProp(_)
             | Expr::Await(_)
-            | Expr::Call(_)
             | Expr::New(..)
+            | Expr::Call(..)
             | Expr::Yield(_)
             | Expr::Assign(_)
             | Expr::PrivateName(_)
@@ -796,6 +888,10 @@ impl Optimizer<'_> {
                         if usage.var_kind.is_none() {
                             return None;
                         }
+
+                        if !usage.reassigned && usage.no_side_effect_for_member_access {
+                            return None;
+                        }
                     }
                 }
                 _ => {}
@@ -818,7 +914,7 @@ impl Optimizer<'_> {
             Expr::Array(arr) => {
                 let mut exprs = vec![];
                 self.changed = true;
-                log::trace!("ignore_return_value: Inverting an array literal");
+                log::debug!("ignore_return_value: Inverting an array literal");
                 for elem in arr.elems.take() {
                     match elem {
                         Some(mut elem) => {
@@ -841,7 +937,7 @@ impl Optimizer<'_> {
             Expr::Object(obj) => {
                 let mut exprs = vec![];
                 self.changed = true;
-                log::trace!("ignore_return_value: Inverting an object literal");
+                log::debug!("ignore_return_value: Inverting an object literal");
                 for prop in obj.props.take() {
                     match prop {
                         PropOrSpread::Spread(mut e) => {
@@ -898,6 +994,7 @@ impl Optimizer<'_> {
             }) if (self.options.negate_iife
                 || self.options.reduce_vars
                 || self.options.side_effects)
+                && !self.ctx.dont_use_negated_iife
                 && match &**arg {
                     Expr::Call(arg) => match &arg.callee {
                         ExprOrSuper::Expr(callee) => match &**callee {
@@ -909,14 +1006,14 @@ impl Optimizer<'_> {
                     _ => false,
                 } =>
             {
-                log::trace!("ignore_return_value: Preserving negated iife");
+                log::debug!("ignore_return_value: Preserving negated iife");
                 return Some(e.take());
             }
 
             // `delete` is handled above
             Expr::Unary(expr) => {
                 self.changed = true;
-                log::trace!("ignore_return_value: Reducing unary ({})", expr.op);
+                log::debug!("ignore_return_value: Reducing unary ({})", expr.op);
                 return self.ignore_return_value(&mut expr.arg);
             }
 
@@ -934,10 +1031,24 @@ impl Optimizer<'_> {
             }
 
             Expr::Cond(cond) => {
+                self.restore_negated_iife(cond);
+
+                let ctx = Ctx {
+                    dont_use_negated_iife: self.ctx.dont_use_negated_iife
+                        || self.options.side_effects,
+                    ..self.ctx
+                };
+
                 let cons_span = cond.cons.span();
                 let alt_span = cond.alt.span();
-                let cons = self.ignore_return_value(&mut cond.cons).map(Box::new);
-                let alt = self.ignore_return_value(&mut cond.alt).map(Box::new);
+                let cons = self
+                    .with_ctx(ctx)
+                    .ignore_return_value(&mut cond.cons)
+                    .map(Box::new);
+                let alt = self
+                    .with_ctx(ctx)
+                    .ignore_return_value(&mut cond.alt)
+                    .map(Box::new);
 
                 // TODO: Remove if test is side effect free.
 
@@ -973,7 +1084,11 @@ impl Optimizer<'_> {
                         if idx == 0 && self.ctx.is_this_aware_callee && is_injected_zero {
                             return Some(*expr.take());
                         }
-                        self.ignore_return_value(&mut **expr)
+                        let ctx = Ctx {
+                            dont_use_negated_iife: idx != 0,
+                            ..self.ctx
+                        };
+                        self.with_ctx(ctx).ignore_return_value(&mut **expr)
                     })
                     .map(Box::new)
                     .collect::<Vec<_>>();
@@ -998,6 +1113,10 @@ impl Optimizer<'_> {
                         }
                     }
 
+                    if exprs.is_empty() {
+                        return None;
+                    }
+
                     return Some(Expr::Seq(SeqExpr {
                         span: seq.span,
                         exprs,
@@ -1082,7 +1201,7 @@ impl Optimizer<'_> {
             })
             .unwrap_or(js_word!(""));
 
-        log::trace!("Converting call to RegExp into a regexp literal");
+        log::debug!("Converting call to RegExp into a regexp literal");
         self.changed = true;
         *e = Expr::Lit(Lit::Regex(Regex {
             span,
@@ -1147,7 +1266,7 @@ impl Optimizer<'_> {
             _ => false,
         }) {
             self.changed = true;
-            log::trace!("Merging variable declarations");
+            log::debug!("Merging variable declarations");
 
             let orig = take(stmts);
             let mut new = Vec::with_capacity(orig.len());
@@ -1195,9 +1314,14 @@ impl Optimizer<'_> {
                 // Remove nested blocks
                 if bs.stmts.len() == 1 {
                     if let Stmt::Block(block) = &mut bs.stmts[0] {
-                        log::trace!("optimizer: Removing nested block");
-                        self.changed = true;
-                        bs.stmts = block.stmts.take();
+                        if block.stmts.iter().all(|stmt| match stmt {
+                            Stmt::Decl(..) => false,
+                            _ => true,
+                        }) {
+                            log::debug!("optimizer: Removing nested block");
+                            self.changed = true;
+                            bs.stmts = block.stmts.take();
+                        }
                     }
                 }
 
@@ -1213,7 +1337,7 @@ impl Optimizer<'_> {
                         _ => false,
                     })
                 {
-                    log::trace!("optimizer: Unwrapping a block with variable statements");
+                    log::debug!("optimizer: Unwrapping a block with variable statements");
                     self.changed = true;
                     *s = bs.stmts[0].take();
                     return;
@@ -1223,7 +1347,7 @@ impl Optimizer<'_> {
                     if let Stmt::Block(block) = &stmt {
                         if block.stmts.is_empty() {
                             self.changed = true;
-                            log::trace!("optimizer: Removing empty block");
+                            log::debug!("optimizer: Removing empty block");
                             *stmt = Stmt::Empty(EmptyStmt { span: DUMMY_SP });
                         }
                     }
@@ -1233,12 +1357,12 @@ impl Optimizer<'_> {
                     match &bs.stmts[0] {
                         Stmt::Expr(..) | Stmt::If(..) => {
                             *s = bs.stmts[0].take();
-                            log::trace!("optimizer: Unwrapping block stmt");
+                            log::debug!("optimizer: Unwrapping block stmt");
                             self.changed = true;
                         }
                         Stmt::Decl(Decl::Fn(..)) if !self.ctx.in_strict => {
                             *s = bs.stmts[0].take();
-                            log::trace!("optimizer: Unwrapping block stmt in non strcit mode");
+                            log::debug!("optimizer: Unwrapping block stmt in non strcit mode");
                             self.changed = true;
                         }
                         _ => {}
@@ -1267,6 +1391,14 @@ impl Optimizer<'_> {
                 self.try_removing_block(&mut s.body, true);
             }
 
+            Stmt::For(s) => {
+                self.try_removing_block(&mut s.body, true);
+            }
+
+            Stmt::ForOf(s) => {
+                self.try_removing_block(&mut s.body, true);
+            }
+
             _ => {}
         }
 
@@ -1303,7 +1435,7 @@ impl Optimizer<'_> {
             match &mut *stmt.cons {
                 Stmt::Expr(cons) => {
                     self.changed = true;
-                    log::trace!("Converting if statement to a form `test && cons`");
+                    log::debug!("Converting if statement to a form `test && cons`");
                     *s = Stmt::Expr(ExprStmt {
                         span: stmt.span,
                         expr: Box::new(stmt.test.take().make_bin(op!("&&"), *cons.expr.take())),
@@ -1345,8 +1477,6 @@ impl VisitMut for Optimizer<'_> {
 
         self.compress_bin_assignment_to_left(e);
         self.compress_bin_assignment_to_right(e);
-
-        self.vars_accessible_without_side_effect.clear();
     }
 
     fn visit_mut_assign_pat_prop(&mut self, n: &mut AssignPatProp) {
@@ -1363,7 +1493,22 @@ impl VisitMut for Optimizer<'_> {
     }
 
     fn visit_mut_bin_expr(&mut self, n: &mut BinExpr) {
-        n.visit_mut_children_with(self);
+        {
+            let ctx = Ctx {
+                in_cond: self.ctx.in_cond
+                    || match n.op {
+                        op!("&&") | op!("||") | op!("??") => true,
+                        _ => false,
+                    },
+                ..self.ctx
+            };
+
+            n.visit_mut_children_with(&mut *self.with_ctx(ctx));
+        }
+
+        self.compress_typeof_undefined(n);
+
+        self.compress_comparsion_of_typeof(n);
 
         self.optimize_bin_operator(n);
 
@@ -1404,29 +1549,48 @@ impl VisitMut for Optimizer<'_> {
     }
 
     fn visit_mut_call_expr(&mut self, e: &mut CallExpr) {
+        let is_this_undefined = match &e.callee {
+            ExprOrSuper::Super(_) => false,
+            ExprOrSuper::Expr(e) => e.is_ident(),
+        };
         {
             let ctx = Ctx {
-                is_this_aware_callee: match &e.callee {
-                    ExprOrSuper::Super(_) => false,
-                    ExprOrSuper::Expr(callee) => is_callee_this_aware(&callee),
-                },
+                is_callee: true,
+                is_this_aware_callee: is_this_undefined
+                    || match &e.callee {
+                        ExprOrSuper::Super(_) => false,
+                        ExprOrSuper::Expr(callee) => is_callee_this_aware(&callee),
+                    },
                 ..self.ctx
             };
             e.callee.visit_mut_with(&mut *self.with_ctx(ctx));
         }
 
-        match &e.callee {
-            ExprOrSuper::Super(_) => {}
-            ExprOrSuper::Expr(e) => {
-                if e.may_have_side_effects() {
-                    self.vars_accessible_without_side_effect.clear();
-                }
+        if is_this_undefined {
+            match &mut e.callee {
+                ExprOrSuper::Expr(callee) => match &mut **callee {
+                    Expr::Member(..) => {
+                        let zero = Box::new(Expr::Lit(Lit::Num(Number {
+                            span: DUMMY_SP,
+                            value: 0.0,
+                        })));
+                        log::trace!("injecting zero to preserve `this` in call");
+
+                        *callee = Box::new(Expr::Seq(SeqExpr {
+                            span: callee.span(),
+                            exprs: vec![zero, callee.take()],
+                        }));
+                    }
+                    _ => {}
+                },
+                _ => {}
             }
         }
 
         {
             let ctx = Ctx {
                 is_this_aware_callee: false,
+                in_call_arg: true,
                 ..self.ctx
             };
             // TODO: Prevent inline if callee is unknown.
@@ -1436,8 +1600,6 @@ impl VisitMut for Optimizer<'_> {
         self.optimize_symbol_call_unsafely(e);
 
         self.inline_args_of_iife(e);
-
-        self.vars_accessible_without_side_effect.clear();
     }
 
     fn visit_mut_class(&mut self, n: &mut Class) {
@@ -1471,6 +1633,8 @@ impl VisitMut for Optimizer<'_> {
     fn visit_mut_cond_expr(&mut self, n: &mut CondExpr) {
         n.visit_mut_children_with(self);
 
+        self.negate_cond_expr(n);
+
         self.optimize_expr_in_bool_ctx(&mut n.test);
     }
 
@@ -1546,15 +1710,6 @@ impl VisitMut for Optimizer<'_> {
 
         self.collapse_seq_exprs(e);
 
-        // Normalize
-        match e {
-            Expr::Paren(paren) => {
-                self.changed = true;
-                *e = *paren.expr.take();
-            }
-            _ => {}
-        }
-
         self.drop_unused_assignments(e);
 
         self.compress_regexp(e);
@@ -1583,7 +1738,7 @@ impl VisitMut for Optimizer<'_> {
             Expr::Bin(bin) => {
                 let expr = self.optimize_lit_cmp(bin);
                 if let Some(expr) = expr {
-                    log::trace!("Optimizing: Literal comparison");
+                    log::debug!("Optimizing: Literal comparison");
                     self.changed = true;
                     *e = expr;
                 }
@@ -1598,14 +1753,14 @@ impl VisitMut for Optimizer<'_> {
         self.handle_negated_seq(e);
         self.compress_array_join(e);
 
-        self.compress_logical_exprs_with_negated_lhs(e);
-
         self.remove_useless_pipes(e);
 
         self.optimize_bools(e);
 
         self.handle_property_access(e);
 
+        self.lift_seqs_of_bin(e);
+
         self.lift_seqs_of_cond_assign(e);
 
         if self.options.negate_iife {
@@ -1619,11 +1774,30 @@ impl VisitMut for Optimizer<'_> {
         self.invoke_iife(e);
 
         self.optimize_bangbang(e);
+
+        match e {
+            Expr::Seq(s) if s.exprs.is_empty() => {
+                e.take();
+            }
+            _ => {}
+        }
     }
 
     fn visit_mut_expr_stmt(&mut self, n: &mut ExprStmt) {
         n.visit_mut_children_with(self);
 
+        let mut need_ignore_return_value = false;
+
+        // If negate_iife is true, it's already handled by
+        // visit_mut_children_with(self) above.
+        if !self.options.negate_iife {
+            // I(kdy1) don't know why this check if required, but there are two test cases
+            // with `options.expressions` as only difference.
+            if !self.options.expr {
+                need_ignore_return_value |= self.negate_iife_in_cond(&mut n.expr);
+            }
+        }
+
         self.negate_iife_ignoring_ret(&mut n.expr);
 
         let is_directive = match &*n.expr {
@@ -1635,9 +1809,18 @@ impl VisitMut for Optimizer<'_> {
             return;
         }
 
-        if self.options.unused
+        if need_ignore_return_value
+            || self.options.unused
             || self.options.side_effects
             || (self.options.sequences() && n.expr.is_seq())
+            || (self.options.conditionals
+                && match &*n.expr {
+                    Expr::Bin(BinExpr {
+                        op: op!("||") | op!("&&"),
+                        ..
+                    }) => true,
+                    _ => false,
+                })
         {
             // Preserve top-level negated iifes.
             match &*n.expr {
@@ -1656,6 +1839,18 @@ impl VisitMut for Optimizer<'_> {
             }
             let expr = self.ignore_return_value(&mut n.expr);
             n.expr = expr.map(Box::new).unwrap_or_else(|| undefined(DUMMY_SP));
+        } else {
+            match &mut *n.expr {
+                Expr::Seq(e) => {
+                    // Non-last items are handled by visit_mut_seq_expr
+                    if let Some(e) = e.exprs.last_mut() {
+                        self.optimize_bang_within_logical_ops(e, true);
+                    }
+                }
+                _ => {
+                    self.optimize_bang_within_logical_ops(&mut *n.expr, true);
+                }
+            }
         }
     }
 
@@ -1685,6 +1880,13 @@ impl VisitMut for Optimizer<'_> {
         n.left.visit_mut_with(&mut *self.with_ctx(ctx));
 
         n.body.visit_mut_with(self);
+
+        match &mut *n.body {
+            Stmt::Block(body) => {
+                self.negate_if_terminate(&mut body.stmts, false, true);
+            }
+            _ => {}
+        }
     }
 
     fn visit_mut_for_of_stmt(&mut self, n: &mut ForOfStmt) {
@@ -1697,22 +1899,42 @@ impl VisitMut for Optimizer<'_> {
         n.left.visit_mut_with(&mut *self.with_ctx(ctx));
 
         n.body.visit_mut_with(self);
+
+        match &mut *n.body {
+            Stmt::Block(body) => {
+                self.negate_if_terminate(&mut body.stmts, false, true);
+            }
+            _ => {}
+        }
     }
 
     fn visit_mut_for_stmt(&mut self, s: &mut ForStmt) {
-        s.visit_mut_children_with(self);
+        let ctx = Ctx {
+            executed_multiple_time: true,
+            ..self.ctx
+        };
 
-        self.optimize_init_of_for_stmt(s);
+        s.visit_mut_children_with(&mut *self.with_ctx(ctx));
 
-        self.drop_if_break(s);
+        self.with_ctx(ctx).merge_for_if_break(s);
+
+        self.with_ctx(ctx).optimize_init_of_for_stmt(s);
+
+        self.with_ctx(ctx).drop_if_break(s);
 
         if let Some(test) = &mut s.test {
-            self.optimize_expr_in_bool_ctx(test);
+            self.with_ctx(ctx).optimize_expr_in_bool_ctx(test);
+        }
+
+        match &mut *s.body {
+            Stmt::Block(body) => {
+                self.negate_if_terminate(&mut body.stmts, false, true);
+            }
+            _ => {}
         }
     }
 
     fn visit_mut_function(&mut self, n: &mut Function) {
-        let old_vars = take(&mut self.vars_accessible_without_side_effect);
         {
             let ctx = Ctx {
                 stmt_lablled: false,
@@ -1721,6 +1943,8 @@ impl VisitMut for Optimizer<'_> {
             n.decorators.visit_mut_with(&mut *self.with_ctx(ctx));
         }
 
+        let old_in_asm = self.ctx.in_asm;
+
         {
             let ctx = Ctx {
                 stmt_lablled: false,
@@ -1736,6 +1960,9 @@ impl VisitMut for Optimizer<'_> {
                 Some(body) => {
                     // Bypass block scope handler.
                     body.visit_mut_children_with(optimizer);
+                    optimizer.remove_useless_return(&mut body.stmts);
+
+                    optimizer.negate_if_terminate(&mut body.stmts, true, false);
 
                     if let Some(last) = body.stmts.last_mut() {
                         optimizer.drop_unused_stmt_at_end_of_fn(last);
@@ -1745,6 +1972,10 @@ impl VisitMut for Optimizer<'_> {
             }
         }
 
+        if let Some(body) = &mut n.body {
+            self.merge_if_returns(&mut body.stmts);
+        }
+
         {
             let ctx = Ctx {
                 can_inline_arguments: true,
@@ -1753,11 +1984,7 @@ impl VisitMut for Optimizer<'_> {
             self.with_ctx(ctx).optimize_usage_of_arguments(n);
         }
 
-        if let Some(body) = &mut n.body {
-            self.merge_if_returns(&mut body.stmts);
-        }
-
-        self.vars_accessible_without_side_effect = old_vars;
+        self.ctx.in_asm = old_in_asm;
     }
 
     fn visit_mut_if_stmt(&mut self, n: &mut IfStmt) {
@@ -1772,7 +1999,13 @@ impl VisitMut for Optimizer<'_> {
 
         n.alt.visit_mut_with(&mut *self.with_ctx(ctx));
 
+        self.negate_if_stmt(n);
+
         self.optimize_expr_in_bool_ctx(&mut n.test);
+
+        self.merge_nested_if(n);
+
+        self.merge_else_if(n);
     }
 
     fn visit_mut_labeled_stmt(&mut self, n: &mut LabeledStmt) {
@@ -1820,17 +2053,55 @@ impl VisitMut for Optimizer<'_> {
     }
 
     fn visit_mut_new_expr(&mut self, n: &mut NewExpr) {
-        n.callee.visit_mut_with(self);
-
-        if n.callee.may_have_side_effects() {
-            self.vars_accessible_without_side_effect.clear();
+        {
+            let ctx = Ctx {
+                is_callee: true,
+                ..self.ctx
+            };
+            n.callee.visit_mut_with(&mut *self.with_ctx(ctx));
         }
 
         {
-            n.args.visit_mut_with(self);
+            let ctx = Ctx {
+                in_call_arg: true,
+                ..self.ctx
+            };
+            n.args.visit_mut_with(&mut *self.with_ctx(ctx));
         }
+    }
 
-        self.vars_accessible_without_side_effect.clear();
+    fn visit_mut_opt_stmt(&mut self, s: &mut Option<Box<Stmt>>) {
+        s.visit_mut_children_with(self);
+
+        match s.as_deref() {
+            Some(Stmt::Empty(..)) => {
+                self.changed = true;
+                log::debug!("misc: Removing empty statement");
+                *s = None;
+            }
+            _ => {}
+        }
+    }
+
+    fn visit_mut_opt_var_decl_or_expr(&mut self, n: &mut Option<VarDeclOrExpr>) {
+        n.visit_mut_children_with(self);
+
+        match n {
+            Some(VarDeclOrExpr::Expr(e)) => match &mut **e {
+                Expr::Seq(SeqExpr { exprs, .. }) if exprs.is_empty() => {
+                    *n = None;
+                    return;
+                }
+                _ => {}
+            },
+            Some(VarDeclOrExpr::VarDecl(v)) => {
+                if v.decls.is_empty() {
+                    *n = None;
+                    return;
+                }
+            }
+            _ => {}
+        }
     }
 
     fn visit_mut_param(&mut self, n: &mut Param) {
@@ -1855,18 +2126,32 @@ impl VisitMut for Optimizer<'_> {
         p.visit_mut_children_with(self);
 
         self.optimize_computed_prop_name_as_normal(p);
+        self.optimize_prop_name(p);
     }
 
     fn visit_mut_return_stmt(&mut self, n: &mut ReturnStmt) {
         n.visit_mut_children_with(self);
 
+        self.drop_undefined_from_return_arg(n);
+
         if let Some(arg) = &mut n.arg {
             self.optimize_in_fn_termiation(&mut **arg);
         }
     }
 
     fn visit_mut_seq_expr(&mut self, n: &mut SeqExpr) {
-        n.visit_mut_children_with(self);
+        {
+            let ctx = Ctx {
+                dont_use_negated_iife: true,
+                ..self.ctx
+            };
+
+            n.visit_mut_children_with(&mut *self.with_ctx(ctx));
+        }
+
+        self.shift_void(n);
+
+        self.shift_assignment(n);
 
         {
             let exprs = n
@@ -1879,6 +2164,7 @@ impl VisitMut for Optimizer<'_> {
                         Expr::Lit(Lit::Num(v)) => v.span.is_dummy(),
                         _ => false,
                     };
+
                     let can_remove =
                         !last && (idx != 0 || !is_injected_zero || !self.ctx.is_this_aware_callee);
 
@@ -1898,16 +2184,19 @@ impl VisitMut for Optimizer<'_> {
             n.exprs = exprs;
         }
 
+        self.merge_sequences_in_seq_expr(n);
+
         self.lift_seqs_of_assign(n);
     }
 
     fn visit_mut_stmt(&mut self, s: &mut Stmt) {
         let ctx = Ctx {
-            in_bang_arg: false,
+            is_callee: false,
             is_delete_arg: false,
-            is_exported: false,
             is_update_arg: false,
             in_lhs_of_assign: false,
+            in_bang_arg: false,
+            is_exported: false,
             in_obj_of_non_computed_member: false,
             ..self.ctx
         };
@@ -1932,7 +2221,7 @@ impl VisitMut for Optimizer<'_> {
                             _ => false,
                         }
                     {
-                        log::trace!("Removing 'use strict'");
+                        log::debug!("Removing 'use strict'");
                         *s = Stmt::Empty(EmptyStmt { span: DUMMY_SP });
                         return;
                     }
@@ -1943,7 +2232,7 @@ impl VisitMut for Optimizer<'_> {
 
                     if can_be_removed {
                         self.changed = true;
-                        log::trace!("Dropping an expression without side effect");
+                        log::debug!("Dropping an expression without side effect");
                         *s = Stmt::Empty(EmptyStmt { span: DUMMY_SP });
                         return;
                     }
@@ -1957,7 +2246,7 @@ impl VisitMut for Optimizer<'_> {
                 Stmt::Debugger(..) => {
                     self.changed = true;
                     *s = Stmt::Empty(EmptyStmt { span: DUMMY_SP });
-                    log::trace!("drop_debugger: Dropped a debugger statement");
+                    log::debug!("drop_debugger: Dropped a debugger statement");
                     return;
                 }
                 _ => {}
@@ -1997,6 +2286,11 @@ impl VisitMut for Optimizer<'_> {
             top_level: false,
             ..self.ctx
         };
+
+        self.with_ctx(ctx).inject_else(stmts);
+
+        self.with_ctx(ctx).merge_sequences_in_stmts(stmts);
+
         self.with_ctx(ctx).handle_stmt_likes(stmts);
 
         self.with_ctx(ctx).merge_var_decls(stmts);
@@ -2021,6 +2315,12 @@ impl VisitMut for Optimizer<'_> {
         }
     }
 
+    fn visit_mut_str(&mut self, s: &mut Str) {
+        s.visit_mut_children_with(self);
+
+        s.kind = Default::default()
+    }
+
     fn visit_mut_switch_cases(&mut self, n: &mut Vec<SwitchCase>) {
         n.visit_mut_children_with(self);
 
@@ -2110,8 +2410,6 @@ impl VisitMut for Optimizer<'_> {
         };
 
         n.visit_mut_children_with(&mut *self.with_ctx(ctx));
-
-        self.vars_accessible_without_side_effect.clear();
     }
 
     fn visit_mut_var_decl(&mut self, n: &mut VarDecl) {
@@ -2131,7 +2429,7 @@ impl VisitMut for Optimizer<'_> {
                 if let Some(e) = &var.init {
                     if is_pure_undefined(e) {
                         self.changed = true;
-                        log::trace!("Dropping explicit initializer which evaluates to `undefined`");
+                        log::debug!("Dropping explicit initializer which evaluates to `undefined`");
 
                         var.init = None;
                     }
@@ -2250,7 +2548,7 @@ fn is_pure_undefined_or_null(e: &Expr) -> bool {
 fn is_callee_this_aware(callee: &Expr) -> bool {
     match &*callee {
         Expr::Arrow(..) => return false,
-        Expr::Seq(s) => return is_callee_this_aware(s.exprs.last().unwrap()),
+        Expr::Seq(..) => return true,
         Expr::Member(MemberExpr {
             obj: ExprOrSuper::Expr(obj),
             ..
