@@ -60,6 +60,10 @@ impl Optimizer<'_> {
             return;
         }
 
+        if self.ctx.is_callee {
+            return;
+        }
+
         if obj.props.iter().any(|prop| match prop {
             PropOrSpread::Spread(_) => false,
             PropOrSpread::Prop(p) => match &**p {
@@ -87,7 +91,7 @@ impl Optimizer<'_> {
                     Prop::Shorthand(_) => {}
                     Prop::KeyValue(p) => {
                         if prop_name_eq(&p.key, &key.sym) {
-                            log::trace!("properties: Inlining a key-value property `{}`", key.sym);
+                            log::debug!("properties: Inlining a key-value property `{}`", key.sym);
                             self.changed = true;
                             *e = *p.value.clone();
                             return;
