@@ -14,6 +14,7 @@
 
 use crate::compress::compressor;
 use crate::hygiene::unique_marker;
+use crate::marks::Marks;
 use crate::option::ExtraOptions;
 use crate::option::MinifyOptions;
 use crate::pass::compute_char_freq::compute_char_freq;
@@ -23,9 +24,13 @@ use crate::pass::hygiene::hygiene_optimizer;
 pub use crate::pass::hygiene::optimize_hygiene;
 use crate::pass::mangle_names::name_mangler;
 use crate::pass::mangle_props::mangle_properties;
-use crate::pass::single::single_pass_optimizer;
+use crate::pass::precompress::precompress_optimizer;
 use analyzer::analyze;
+use pass::postcompress::postcompress_optimizer;
+use std::time::Instant;
 use swc_common::comments::Comments;
+use swc_common::sync::Lrc;
+use swc_common::SourceMap;
 use swc_ecma_ast::Module;
 use swc_ecma_visit::FoldWith;
 use swc_ecma_visit::VisitMutWith;
@@ -35,6 +40,7 @@ mod analyzer;
 mod compress;
 mod debug;
 mod hygiene;
+mod marks;
 pub mod option;
 mod pass;
 pub mod timing;
@@ -43,11 +49,15 @@ mod util;
 #[inline]
 pub fn optimize(
     mut m: Module,
+    cm: Lrc<SourceMap>,
     comments: Option<&dyn Comments>,
     mut timings: Option<&mut Timings>,
     options: &MinifyOptions,
     extra: &ExtraOptions,
 ) -> Module {
+    let marks = Marks::new();
+
+    let start = Instant::now();
     if let Some(defs) = options.compress.as_ref().map(|c| &c.global_defs) {
         // Apply global defs.
         //
@@ -60,10 +70,13 @@ pub fn optimize(
             m.visit_mut_with(&mut global_defs::globals_defs(defs, extra.top_level_mark));
         }
     }
+    log::info!("global_defs took {:?}", Instant::now() - start);
 
-    m.visit_mut_with(&mut single_pass_optimizer(
-        options.compress.clone().unwrap_or_default(),
-    ));
+    if let Some(options) = &options.compress {
+        let start = Instant::now();
+        m.visit_mut_with(&mut precompress_optimizer(options.clone()));
+        log::info!("precompress took {:?}", Instant::now() - start);
+    }
 
     m.visit_mut_with(&mut unique_marker());
 
@@ -95,8 +108,14 @@ pub fn optimize(
         t.section("compress");
     }
     if let Some(options) = &options.compress {
-        m = m.fold_with(&mut compressor(&options, comments));
+        let start = Instant::now();
+        m = m.fold_with(&mut compressor(cm.clone(), marks, &options, comments));
+        log::info!("compressor took {:?}", Instant::now() - start);
         // Again, we don't need to validate ast
+
+        let start = Instant::now();
+        m.visit_mut_with(&mut postcompress_optimizer(options));
+        log::info!("postcompressor took {:?}", Instant::now() - start);
     }
 
     if let Some(ref mut _t) = timings {
