@@ -28,9 +28,12 @@ pub fn name_mangler(options: MangleOptions, _char_freq_info: CharFreqInfo) -> im
 struct Mangler {
     options: MangleOptions,
     n: usize,
+    private_n: usize,
+
     preserved: FxHashSet<Id>,
     preserved_symbols: FxHashSet<JsWord>,
     renamed: FxHashMap<Id, JsWord>,
+    renamed_private: FxHashMap<Id, JsWord>,
     data: Option<ProgramData>,
 }
 
@@ -65,6 +68,23 @@ impl Mangler {
             break;
         }
     }
+
+    fn rename_private(&mut self, private_name: &mut PrivateName) {
+        let id = private_name.id.to_id();
+
+        let new_sym = if let Some(cached) = self.renamed_private.get(&id) {
+            cached.clone()
+        } else {
+            let sym: JsWord = base54(self.private_n).into();
+            self.private_n += 1;
+
+            self.renamed_private.insert(id.clone(), sym.clone());
+
+            sym
+        };
+
+        private_name.id.sym = new_sym;
+    }
 }
 
 impl VisitMut for Mangler {
@@ -75,7 +95,6 @@ impl VisitMut for Mangler {
 
         n.class.visit_mut_with(self);
     }
-
     fn visit_mut_export_named_specifier(&mut self, n: &mut ExportNamedSpecifier) {
         if n.exported.is_none() {
             n.exported = Some(n.orig.clone());
@@ -83,7 +102,6 @@ impl VisitMut for Mangler {
 
         self.rename(&mut n.orig);
     }
-
     fn visit_mut_expr(&mut self, e: &mut Expr) {
         e.visit_mut_children_with(self);
 
@@ -100,6 +118,25 @@ impl VisitMut for Mangler {
         n.function.visit_mut_with(self);
     }
 
+    fn visit_mut_import_default_specifier(&mut self, n: &mut ImportDefaultSpecifier) {
+        self.rename(&mut n.local);
+    }
+
+    fn visit_mut_import_named_specifier(&mut self, n: &mut ImportNamedSpecifier) {
+        match &n.imported {
+            Some(..) => {}
+            None => {
+                n.imported = Some(n.local.clone());
+            }
+        }
+
+        self.rename(&mut n.local);
+    }
+
+    fn visit_mut_import_star_as_specifier(&mut self, n: &mut ImportStarAsSpecifier) {
+        self.rename(&mut n.local);
+    }
+
     fn visit_mut_labeled_stmt(&mut self, n: &mut LabeledStmt) {
         n.body.visit_mut_with(self);
     }
@@ -138,6 +175,12 @@ impl VisitMut for Mangler {
         }
     }
 
+    fn visit_mut_private_name(&mut self, private_name: &mut PrivateName) {
+        if !self.options.keep_private_props {
+            self.rename_private(private_name);
+        }
+    }
+
     fn visit_mut_script(&mut self, n: &mut Script) {
         let data = analyze(&*n);
         self.data = Some(data);
